"""
Add the common input arguments
"""

from agent_search.llm.global_variables import (
    local_model_list,
    openai_chat_model_list,
    openai_nonchat_model_list,
    together_model_list,
)


def add_pipeline_arguments(parser):
    # determine components of the pipeline
    parser.add_argument(
        "--pipeline",
        type=str,
        required=True,
        choices=[
            "generate_and_correct",
            "retrieve_and_generate",
            "generate",
            "retrieve_only",
            "early_combine",
            "atlas",
        ],
        default="generate_and_correct",
        help="The type of pipeline used to imrpove GPT-3 response. Only used to know which modules to load.",
    )
    parser.add_argument(
        "--claim_prompt_template_file",
        type=str,
        default="split_claims.prompt",
        help="The path to the file containing the claim LLM prompt.",
    )
    parser.add_argument(
        "--refinement_prompt",
        default="refine_w_feedback.prompt",
        help="What prompt to use to refine the final response.",
    )
    parser.add_argument(
        "--do_refine",
        action="store_true",
        help="Whether to refine the final response.",
    )
    parser.add_argument(
        "--skip_verification",
        action="store_true",
        help="If True, all claims will be considered correct without fact-checking. Especially useful to speed up debugging of the other parts of the pipeline.",
    )

    parser.add_argument(
        "--fuse_claim_splitting",
        action="store_true",
        help="If True, The first claim splitting stage of early_combine pipeline will be fused with the generate stage. Only useful for distilled models that have been trained to do this.",
    )

    parser.add_argument(
        "--colbert_endpoint",
        type=str,
        default="http://127.0.0.1:5000/search",
        help="whether using colbert for retrieval.",
    )
    parser.add_argument(
        "--engine",
        type=str,
        required=True,
        choices=["atlas"]
        + local_model_list
        + together_model_list
        + openai_chat_model_list
        + openai_nonchat_model_list,
        help="The LLM engine to use.",
    )  # choices are from the smallest to the largest model

    parser.add_argument(
        "--generate_engine",
        type=str,
        default=None,
        choices=["atlas"]
        + local_model_list
        + together_model_list
        + openai_chat_model_list
        + openai_nonchat_model_list,
        help="The LLM engine to use for the 'generate' stage of pipelines. If provided, overrides --engine for that stage.",
    )  # choices are from the smallest to the largest model

    parser.add_argument(
        "--draft_engine",
        type=str,
        default=None,
        choices=["atlas"]
        + local_model_list
        + together_model_list
        + openai_chat_model_list
        + openai_nonchat_model_list,
        help="The LLM engine to use for the 'draft' stage of pipelines. If provided, overrides --engine for that stage.",
    )  # choices are from the smallest to the largest model

    parser.add_argument(
        "--reranking_method",
        type=str,
        choices=["none", "date"],
        default="none",
        help="Only used for retrieve_and_generate pipeline",
    )

    # LLM generation hyperparameters
    parser.add_argument(
        "--max_tokens",
        type=int,
        default=250,
        required=False,
        help="Only affects user-facing prompts",
    )
    parser.add_argument(
        "--temperature",
        type=float,
        default=0.8,
        required=False,
        help="Only affects user-facing prompts",
    )
    parser.add_argument(
        "--top_p",
        type=float,
        default=0.9,
        required=False,
        help="Only affects user-facing prompts",
    )
    parser.add_argument(
        "--frequency_penalty",
        type=float,
        default=0.0,
        required=False,
        help="Only affects user-facing prompts",
    )
    parser.add_argument(
        "--presence_penalty",
        type=float,
        default=0.0,
        required=False,
        help="Only affects user-facing prompts",
    )

    parser.add_argument(
        "--evi_num",
        type=int,
        default=2,
        help="Number of evidences to retrieve per claim.",
    )

    parser.add_argument(
        "--retrieval_num",
        type=int,
        default=3,
        help="Number of passages to retrieve when searching for information.",
    )


def check_pipeline_arguments(args):
    # make sure for ATLAS, both engine and pipeline are set to 'atlas'
    if hasattr(args, "pipeline"):
        if (args.engine == "atlas" and args.pipeline != "atlas") or (
            args.engine != "atlas" and args.pipeline == "atlas"
        ):
            raise ValueError(
                "When using ATLAS, both `engine` and `pipeline` input arguments should be set to 'atlas'."
            )



import logging
from typing import List

from .dialog_turn import DialogueTurn

# from agent_search.llm.llm_generate import llm_generate

logger = logging.getLogger(__name__)


class Refiner:
    def __init__(self, prompt, args):
        self.prompt = prompt
        self.temperature = args.temperature
        self.top_p = args.top_p

    def set_refinement_fields(
        self,
        object_dlg_history: List[DialogueTurn],
        new_dlg_turn: DialogueTurn,
        engine_dict,
    ):
        prompt_output = llm_generate(
            template_file=self.prompt,
            prompt_parameter_values={
                "dlg": object_dlg_history,
                "new_dlg_turn": new_dlg_turn,
            },
            engine=engine_dict["default"],
            max_tokens=300,
            temperature=self.temperature,
            top_p=self.top_p,
            stop_tokens=None,
            postprocess=False,
        )
        if self.prompt.endswith("refine_w_feedback.prompt"):
            return Refiner.handle_refinement_with_feedback(
                new_dlg_turn, prompt_output
            )
        elif self.prompt.endswith("refine.prompt"):
            return Refiner.handle_refinement_without_feedback(
                new_dlg_turn, prompt_output
            )
        else:
            raise ValueError("Unknown refinement prompt.")

    @staticmethod
    def handle_refinement_without_feedback(new_dlg_turn, prompt_output):
        new_dlg_turn.refined_utterance = prompt_output.strip()
        return new_dlg_turn.refined_utterance

    @staticmethod
    def handle_refinement_with_feedback(new_dlg_turn, prompt_output: str):
        refine_identifiers = [
            "Revised response after applying this feedback:",
            "Response after applying this feedback:",
        ]
        for identifier in refine_identifiers:
            if identifier in prompt_output:
                feedback, prompt_output = prompt_output.split(identifier)

                (
                    new_dlg_turn.feedback,
                    new_dlg_turn.feedback_scores,
                ) = Refiner._parse_feedback(feedback)
                if sum(new_dlg_turn.feedback_scores) == 100 * len(
                    new_dlg_turn.feedback_scores
                ):
                    # skip refinement if it already gets full score
                    new_dlg_turn.refined_utterance = (
                        new_dlg_turn.agent_utterance
                    )
                else:
                    new_dlg_turn.refined_utterance = prompt_output.strip()
                return new_dlg_turn.refined_utterance

        logger.error(
            "Skipping refinement due to malformatted Refined response: %s",
            prompt_output,
        )
        new_dlg_turn.refined_utterance = new_dlg_turn.agent_utterance
        return new_dlg_turn.refined_utterance

    @staticmethod
    def _parse_feedback(feedback):
        if "User:" in feedback:
            feedback = feedback.split("User:")[0]
        feedback_lines = feedback.strip().split("\n")

        if len(feedback_lines) < 4 or len(feedback_lines) > 5:
            logger.error("Feedback malformatted")
            logger.error(feedback_lines)
            return [], []

        scores = (
            []
        )  # Relevant, Informative, Conversational, Non-Redundant, Temporally Correct scores
        for line in feedback_lines:
            score = line.strip().split(" ")[-1].strip()
            if (
                score == "N/A" or "this criterion is not applicable" in line
            ):  # some models say "not applicable" instead of N/A
                score = 100
            else:
                try:
                    score = int(score.split("/")[0])
                except:
                    logger.error(f"Feedback line malformatted: {line}")
                    score = 100
            scores.append(score)
        logger.info("Feedback scores: %s", scores)
        return feedback_lines, scores



import logging
from typing import List

from agent_search.llm.llm_generate import llm_generate

logger = logging.getLogger(__name__)


class ClaimSplitter:
    def __init__(self, prompt_template_file: str):
        self.prompt_template_file = prompt_template_file

    def split_claim(
        self,
        dialog_history: List,
        new_user_utterance: str,
        current_agent_utterance: str,
        engine_dict: dict,
        dialog_topic: str = None,
        claims_output=None,
    ):
        """
        dialog_topic: used for splitting claims of a simulated dialog we want to evaluate
        claims_output: If provided, won't run LLM again and will use this output instead
        """
        if not claims_output:
            assert (
                current_agent_utterance is not None
            ), "current_agent_utterance must be provided when `split_claim` is not fused with `generate`"
            claims_output = llm_generate(
                template_file=self.prompt_template_file,
                prompt_parameter_values={
                    "dlg": dialog_history,
                    "new_user_utterance": new_user_utterance,
                    "current_agent_utterance": current_agent_utterance,
                    "dialog_topic": dialog_topic,
                },
                engine=engine_dict["default"],
                max_tokens=300,
                temperature=0,
                stop_tokens=["====="],
                postprocess=False,
            )

        if claims_output.startswith("Yes. "):
            # necessary for some versions of distilled models
            claims_output = claims_output[5:]
        all_claims = self._format_output(claims_output)

        return all_claims

    def _format_output(self, output):
        lines = output.split("\n")
        if lines[0].startswith("Nothing."):
            # no claims detected
            return []
        all_claims = []
        try:
            for c in lines:
                claim = c
                cleaned_claim = claim.split(f"- ")[-1].strip()
                if cleaned_claim:
                    split_term = " The year of the results is "
                    if split_term not in cleaned_claim:
                        split_term = " The year of the claim is "
                    splitted = cleaned_claim.split(split_term)
                    if len(splitted) == 2:
                        cleaned_claim, year = splitted
                        year = year[1:-2]
                    else:
                        # sometimes model output may not be well-formatted (e.g. N/A); default to none
                        cleaned_claim = splitted[0]
                        year = "none"
                    all_claims.append((cleaned_claim, year))
        except Exception as e:
            logger.error(
                "Error while parsing claims in %s: %s", output, str(e)
            )
            raise e

        return all_claims

    @staticmethod
    def remove_claims_from_previous_turns(claims: List, object_dlg_history):
        """
        Removes claims that are repeated from the last turn. This is often the result of LLM making a mistake while splitting claims.
        But even if it is actually a claim that the chatbot repeats, removing it here is beneficial as it will reduce repetitiveness.
        """
        previous_turn_claims = []
        for i in range(len(object_dlg_history)):
            previous_turn_claims.extend(
                [c[0] for c in object_dlg_history[i].claims]
            )
        claims = [c for c in claims if c[0] not in previous_turn_claims]

        return claims



import logging
import os
import pathlib
import re

import spacy

logger = logging.getLogger(__name__)

spacy_nlp = spacy.load("en_core_web_sm")


class bcolors:
    HEADER = "\033[95m"
    OKBLUE = "\033[94m"
    OKCYAN = "\033[96m"
    OKGREEN = "\033[92m"
    WARNING = "\033[93m"
    FAIL = "\033[91m"
    ENDC = "\033[0m"
    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"


def print_chatbot(s: str):
    print(bcolors.OKGREEN + bcolors.BOLD + s + bcolors.ENDC)


def input_user() -> str:
    try:
        user_utterance = input(bcolors.OKCYAN + bcolors.BOLD + "User: ")
        # ignore empty inputs
        while not user_utterance.strip():
            user_utterance = input(bcolors.OKCYAN + bcolors.BOLD + "User: ")
    finally:
        print(bcolors.ENDC)
    return user_utterance


def make_parent_directories(file_name: str):
    """
    Creates the parent directories of `file_name` if they don't exist
    """
    pathlib.Path(os.path.dirname(file_name)).mkdir(parents=True, exist_ok=True)


def is_everything_verified(ver_out):
    """
    Everything is verified when 1) we have only one claim and it is supported or 2) all claims are supported.
    """
    for label_fix in ver_out:
        if label_fix["label"] != "SUPPORTS":
            return False
    return True


def extract_year(title, passage):
    if title:
        passage = title + " | " + passage
    years = []
    year_pattern = r"\d{4}"
    year_duration_pattern = r"\b\d{4}[--–]\d{2}\b"
    year_to_pattern = r"\b\d{4} to \d{4}\b"
    # extract "1990 to 1998" before spacy because spacy would split it to 1990 and 1998
    re_year_tos = re.findall(year_to_pattern, passage)
    for re_year_to in re_year_tos:
        re_years = re.findall(year_pattern, re_year_to)
        if len(re_years) != 2:
            continue
        year1, year2 = re_years
        years.extend(list(range(int(year1), int(year2) + 1)))
        passage.replace(re_year_to, " ")

    doc = spacy_nlp(passage)
    dates = [(X.text, X.label_) for X in doc.ents if X.label_ == "DATE"]
    for date in dates:
        date = date[0]
        # "the 2006–07 season"
        re_year_durations = re.findall(year_duration_pattern, date)
        if re_year_durations:
            for re_year_duration in re_year_durations:
                if "–" in re_year_duration:
                    year1, year2 = re_year_duration.split("–")
                elif "-" in re_year_duration:
                    year1, year2 = re_year_duration.split("-")
                else:
                    continue
                year2 = year1[:2] + year2
                years.extend([year1, year2])
            continue
        # any 4 digits
        re_years = re.findall(year_pattern, date)
        if re_years:
            years.extend(re_years)
    years = list(sorted(set([int(year) for year in years])))
    return years



import os
import threading

import yaml
from jinja2 import (
    Environment,
    FileSystemLoader,
    StrictUndefined,
    select_autoescape,
)

from agent_search.llm.tgi_client import TGIOpenAILikeClient

# singleton
jinja_environment = Environment(
    loader=FileSystemLoader(["./", "./pipelines/", "./pipelines/prompts/"]),
    autoescape=select_autoescape(),
    trim_blocks=True,
    lstrip_blocks=True,
    line_comment_prefix="#",
    undefined=StrictUndefined,
)  # StrictUndefined raises an exception if a variable in template is not passed to render(), instead of setting it as empty

# define chat tags for chat models
system_start = "<|system_start|>"
system_end = "<|system_end|>"
user_start = "<|user_start|>"
user_end = "<|user_end|>"
assistant_start = "<|assistant_start|>"
assistant_end = "<|assistant_end|>"


llm_config = {}
with open("llm_config.yaml", "r") as config_file:
    llm_config = yaml.unsafe_load(config_file)
    # print(llm_config)
all_llm_endpoints = llm_config["llm_endpoints"]
for a in all_llm_endpoints:
    if "api_key" in a:
        a["api_key"] = os.getenv(a["api_key"])

all_llm_endpoints = [
    a
    for a in all_llm_endpoints
    if "api_key" not in a
    or (a["api_key"] is not None and len(a["api_key"]) > 0)
]  # remove resources for which we don't have a key
# print("all_llm_endpoints = ", all_llm_endpoints)

# for prompt debugging
prompt_log_file = "data/prompt_logs.json"
debug_prompts = False
prompt_logs = []
prompts_to_skip_for_debugging = [
    "benchmark/prompts/user_with_passage.prompt",
    "benchmark/prompts/user_with_topic.prompt",
]

# for local engines
local_engine_clients = {}
for endpoint_idx in range(len(all_llm_endpoints)):
    resource = all_llm_endpoints[endpoint_idx]
    if resource["api_type"] == "local":
        assert (
            "api_base" in resource and "prompt_format" in resource
        ), "Local engines should have api_base and prompt_format"
        local_engine_clients[resource["api_base"]] = TGIOpenAILikeClient(
            timeout=10,
            base_url=resource["api_base"],
        )


def set_debug_mode():
    global debug_prompts
    debug_prompts = True


# this code is not safe to use with multiprocessing, only multithreading
thread_lock = threading.Lock()

local_model_list = ["local"]
together_model_list = [
    "RedPajama-INCITE-7B-Instruct",
    "llama-2-70b",
    "llama-2-13b",
    "llama-2-7b",
    "llama-2-7b-32k",
    "mixtral-8-7b",
    "mistral-7b",
]
openai_chat_model_list = ["gpt-35-turbo", "gpt-4-32k", "gpt-4"]
openai_nonchat_model_list = [
    "gpt-35-turbo-instruct",
    "text-davinci-002",
    "text-davinci-003",
]

inference_cost_per_1000_tokens = {
    # from https://openai.com/pricing
    "ada": (0.0004, 0.0004),
    "babbage": (0.0005, 0.0005),
    "curie": (0.002, 0.002),
    "davinci": (0.02, 0.02),
    "gpt-35-turbo": (0.0015, 0.002),
    "gpt-4-32k": (0.06, 0.12),
    "gpt-4": (0.03, 0.06),
    # from https://www.together.ai/pricing
    "llama-2-70b": (0.0009, 0.0009),
    "llama-2-13b": (0.000225, 0.000225),
    "llama-2-7b": (0.0002, 0.0002),
    "llama-2-7b-32k": (0.0002, 0.0002),
    "RedPajama-INCITE-7B-Instruct": (0.0002, 0.0002),
    "mixtral-8-7b": (0.0006, 0.0006),
    "mistral-7b": (0.0002, 0.0002),
}  # (prompt, completion)

total_cost = 0  # in USD


def add_to_total_cost(amount: float):
    global total_cost
    with thread_lock:
        total_cost += amount


def get_total_cost():
    global total_cost
    return total_cost


def _model_name_to_cost(model_name: str) -> float:
    if model_name in local_model_list:
        return 0, 0
    for model_family in inference_cost_per_1000_tokens.keys():
        if model_family in model_name:
            return inference_cost_per_1000_tokens[model_family]
    raise ValueError("Did not recognize OpenAI model name %s" % model_name)



from datetime import datetime
from typing import List

import pytz

from .global_variables import (
    assistant_end,
    assistant_start,
    jinja_environment,
    local_model_list,
    openai_chat_model_list,
    system_end,
    system_start,
    user_end,
    user_start,
)


def _remove_starting_and_ending_whitespace(text):
    # remove whitespace at the beginning and end of each line
    return "\n".join([line.strip() for line in text.split("\n")])


def _remove_chat_tags(s: str):
    return _remove_starting_and_ending_whitespace(
        s.replace(system_start, "")
        .replace(system_end, "")
        .replace(user_start, "")
        .replace(user_end, "")
        .replace(assistant_start, "")
        .replace(assistant_end, "")
    )  # need to remove starting and ending whitespace because after removing chat tags, whitespace that used to be in the middle might become starting or ending whitespace


def _fill_template(
    template_file, prompt_parameter_values, get_rendered_blocks=False
):
    template = jinja_environment.get_template(template_file)

    prompt_parameter_values["instruction_start"] = system_start
    prompt_parameter_values["instruction_end"] = system_end
    prompt_parameter_values["input_start"] = user_start
    prompt_parameter_values["input_end"] = user_end
    prompt_parameter_values["output_start"] = assistant_start
    prompt_parameter_values["output_end"] = assistant_end

    # always make these useful constants available in a template
    # make a new function call each time since the date might change during a long-term server deployment
    today = datetime.now(pytz.timezone("US/Pacific")).date()
    prompt_parameter_values["today"] = today.strftime(
        "%B %d, %Y"
    )  # May 30, 2023
    prompt_parameter_values["current_year"] = today.year
    prompt_parameter_values["location"] = "the U.S."
    prompt_parameter_values["chatbot_name"] = "WikiChat"

    filled_prompt = template.render(**prompt_parameter_values)
    filled_prompt = _remove_starting_and_ending_whitespace(filled_prompt)

    # Access the 'content' block and render it
    rendered_blocks = {}
    if get_rendered_blocks:
        for block_name in template.blocks.keys():
            block = template.blocks[block_name](
                template.new_context(vars=prompt_parameter_values)
            )
            rendered = "".join(block)
            rendered = _remove_chat_tags(
                rendered
            )  # blocks are used for logging and local engines, so should never have chat tags
            rendered_blocks[block_name] = rendered

    return filled_prompt, rendered_blocks


def _fill_prompt(
    template_file, prompt_parameter_values, engine, prompt_format
):
    filled_prompt = [
        _fill_template(template_file, p, get_rendered_blocks=True)
        for p in prompt_parameter_values
    ]
    # convert list of tuples to tuple of lists
    filled_prompt, rendered_blocks = tuple(zip(*filled_prompt))
    filled_prompt, rendered_blocks = list(filled_prompt), list(rendered_blocks)

    # remove short_instruction blocks
    for idx, block in enumerate(rendered_blocks):
        if "short_instruction" in block:
            filled_prompt[idx] = filled_prompt[idx].replace(
                block["short_instruction"], ""
            )

    assert (
        engine in local_model_list or prompt_format == "none"
    ), "Only local engines can have any prompt formats"
    if engine in local_model_list:
        if prompt_format == "none":
            pass
        elif prompt_format == "alpaca":
            # overwrite filled_prompt
            filled_prompt = [
                (
                    "Below is an instruction that describes a task, paired with an input that provides further context. "
                    "Write a response that appropriately completes the request.\n\n"
                    "### Instruction:\n{short_instruction}\n\n### Input:\n{input}\n\n### Response:"
                ).format_map(block)
                for block in rendered_blocks
            ]
        elif prompt_format == "simple":
            # overwrite filled_prompt
            filled_prompt = [
                "{short_instruction}\n\n{input}\n".format_map(block)
                for block in rendered_blocks
            ]
        else:
            raise ValueError(
                "Unknown prompt format specified for the local engine."
            )

    if engine not in openai_chat_model_list:
        # print("rendered_blocks = ", rendered_blocks)
        filled_prompt = [_remove_chat_tags(f) for f in filled_prompt]

    return filled_prompt, rendered_blocks


def _convert_filled_prompt_to_chat_messages(filled_prompt: List[str]):
    """
    ChatGPT and GPT-4 use ChatCompletion and a different format from older models.
    The following is a naive implementation of few-shot prompting, which may be improved.
    """

    ret = []
    # print("before conversion", json.dumps(filled_prompt, indent=2))
    for fp in filled_prompt:
        # TODO check that start system is unique
        messages = []

        system_s = fp.find(system_start)
        system_e = fp.find(system_end, system_s)
        if system_s < 0:
            # did not find a system message in the prompt, so will put everything inside system for backward compatibility
            messages.append(
                {
                    "role": "system",
                    "content": fp.strip(),
                }
            )
            ret.append(messages)
            continue

        messages.append(
            {
                "role": "system",
                "content": fp[system_s + len(system_start) : system_e].strip(),
            }
        )

        last_index = 0
        while True:
            user_s = fp.find(user_start, last_index)
            assistant_s = fp.find(assistant_start, last_index)
            if (
                user_s >= 0
                and assistant_s < 0
                or (user_s >= 0 and user_s < assistant_s)
            ):
                user_e = fp.find(user_end, user_s)
                assert user_e >= 0, "Missing closing tag for user"
                last_index = user_e
                messages.append(
                    {
                        "role": "user",
                        "content": fp[
                            user_s + len(user_start) : user_e
                        ].strip(),
                    }
                )
            elif (
                user_s < 0
                and assistant_s >= 0
                or (assistant_s >= 0 and user_s > assistant_s)
            ):
                assistant_e = fp.find(assistant_end, assistant_s)
                assert assistant_e >= 0, "Missing closing tag for assistant"
                last_index = assistant_e
                messages.append(
                    {
                        "role": "assistant",
                        "content": fp[
                            assistant_s + len(assistant_start) : assistant_e
                        ].strip(),
                    }
                )
            else:
                assert user_s < 0 and assistant_s < 0
                break
        ret.append(messages)

    return ret



import logging
from concurrent.futures import ThreadPoolExecutor

from text_generation import Client

logger = logging.getLogger(__name__)


class TGIOpenAILikeClient(Client):
    """
    Used for local LLMs served via HuggingFace's text-generation-inference
    """

    def __init__(
        self,
        base_url,
        headers=None,
        cookies=None,
        timeout=10,
    ):
        super().__init__(base_url, headers, cookies, timeout)

    def _generate(
        self,
        prompt,
        max_tokens,
        stop,
        temperature,
        repetition_penalty=None,
        top_k=None,
        top_p=None,
    ):
        if temperature == 0:
            temperature = 1
            do_sample = False
        else:
            do_sample = True

        if top_p == 1:
            top_p = None

        return super().generate(
            prompt,
            do_sample,
            max_tokens,
            best_of=None,
            repetition_penalty=repetition_penalty,
            return_full_text=False,
            seed=None,
            stop_sequences=stop,
            temperature=temperature,
            top_k=top_k,
            top_p=top_p,
            truncate=None,
            typical_p=None,
            watermark=False,
            decoder_input_details=False,
        )

    def completion(self, kwargs):
        max_tokens = kwargs["max_tokens"]
        temperature = kwargs["temperature"]
        top_p = kwargs["top_p"]
        frequency_penalty = kwargs["frequency_penalty"]
        presence_penalty = kwargs["presence_penalty"]
        stop = kwargs["stop"]

        if (
            "presence_penalty" in kwargs
            and kwargs["presence_penalty"] is not None
            and kwargs["presence_penalty"] != 0
        ):
            logger.warning(
                "Ignoring `presence_penalty` since it is not supported by this model."
            )
        if (
            "frequency_penalty" in kwargs
            and kwargs["frequency_penalty"] is not None
            and kwargs["frequency_penalty"] != 0
        ):
            logger.warning(
                "Ignoring `frequency_penalty` since it is not supported by this model."
            )

        with ThreadPoolExecutor(len(kwargs["prompt"])) as executor:
            thread_outputs = [
                executor.submit(
                    self._generate,
                    prompt=p,
                    temperature=temperature,
                    max_tokens=max_tokens,
                    top_p=top_p,
                    stop=stop,
                )
                for p in kwargs["prompt"]
            ]
        thread_outputs = [o.result().generated_text for o in thread_outputs]
        # print(thread_outputs)
        ret = {
            "choices": [
                {
                    "finish_reason": "stop",
                    "index": i,
                    "logprobs": None,
                    "text": thread_outputs[i],
                }
                for i in range(len(thread_outputs))
            ],
            "created": 0,
            "id": "",
            "model": kwargs["engine"],
            "object": "text_completion",
            "usage": {
                "completion_tokens": 0,
                "prompt_tokens": 0,
                "total_tokens": 0,
            },
        }
        # print("ret = ", ret)
        return ret



"""
Functionality to work with .prompt files
"""

import asyncio
import copy
import json
import logging
from concurrent.futures import ThreadPoolExecutor
from functools import lru_cache, partial, wraps
from typing import List, Union

import openai
import together
from openai import OpenAIError
from tenacity import (
    before_sleep_log,
    retry,
    retry_if_exception_type,
    stop_after_attempt,
    wait_exponential,
)

import agent_search.llm.global_variables as global_variables
from agent_search.llm.load_prompt import (
    _convert_filled_prompt_to_chat_messages,
    _fill_prompt,
)

logger = logging.getLogger(__name__)


def write_prompt_logs_to_file():
    with open(global_variables.prompt_log_file, "w") as f:
        f.write(
            json.dumps(
                global_variables.prompt_logs, indent=4, ensure_ascii=False
            )
        )


def remove_malformed_prompt_logs(is_malformed_function):
    new_prompt_logs = []
    for item in global_variables.prompt_logs:
        if not is_malformed_function(item):
            new_prompt_logs.append(item)

    global_variables.prompt_logs = new_prompt_logs


def lru_cache_if_zero_temperature(maxsize):
    def decorator(func):
        cached_func = lru_cache(maxsize=maxsize)(func)

        @wraps(func)
        def wrapper(original_engine_name, **kwargs):
            if kwargs.get("temperature") == 0:
                # Convert the inputs to be hashable. Necessary for caching to work.
                for a in kwargs:
                    if isinstance(kwargs[a], list):
                        kwargs[a] = tuple(kwargs[a])
                return cached_func(original_engine_name, **kwargs)
            else:
                return func(original_engine_name, **kwargs)

        return wrapper

    return decorator


async def openai_chat_completion(kwargs):
    engine = kwargs["engine"]
    prompt = kwargs["prompt"]
    max_tokens = kwargs["max_tokens"]
    temperature = kwargs["temperature"]
    top_p = kwargs["top_p"]
    frequency_penalty = kwargs["frequency_penalty"]
    presence_penalty = kwargs["presence_penalty"]
    stop = kwargs["stop"]
    logit_bias = kwargs.get("logit_bias", {})

    if kwargs["api_type"] == "azure":
        f = partial(
            openai.ChatCompletion.acreate,
            engine=engine,
            max_tokens=max_tokens,
            temperature=temperature,
            top_p=top_p,
            frequency_penalty=frequency_penalty,
            presence_penalty=presence_penalty,
            stop=stop,
            logit_bias=logit_bias,
            api_base=kwargs["api_base"],
            api_key=kwargs["api_key"],
            api_type=kwargs["api_type"],
            api_version=kwargs["api_version"],
        )
    else:
        # openai.com
        # use 'model' instead of 'engine'. no `api_version``
        f = partial(
            openai.ChatCompletion.acreate,
            model=engine,
            max_tokens=max_tokens,
            temperature=temperature,
            top_p=top_p,
            frequency_penalty=frequency_penalty,
            presence_penalty=presence_penalty,
            stop=stop,
            logit_bias=logit_bias,
            api_base=kwargs["api_base"],
            api_key=kwargs["api_key"],
            api_type=kwargs["api_type"],
        )

    prompt = _convert_filled_prompt_to_chat_messages(prompt)

    coroutines = [f(messages=p) for p in prompt]
    outputs = await asyncio.gather(*coroutines)

    # make the outputs of these multiple calls to openai look like one call with batching
    ret = outputs[0]
    for i, o in enumerate(outputs[1:]):
        o["choices"][0]["index"] = i + 1
        ret["choices"].append(o["choices"][0])
        ret["usage"]["completion_tokens"] += o["usage"]["completion_tokens"]
        ret["usage"]["prompt_tokens"] += o["usage"]["prompt_tokens"]
        ret["usage"]["total_tokens"] += o["usage"]["total_tokens"]

    for c in ret["choices"]:
        # key 'content' does not exist for gpt-4 when output is empty.
        if c["finish_reason"] == "content_filter":
            c["text"] = None
        else:
            c["text"] = c["message"]["content"]

    return ret


def openai_nonchat_completion(kwargs):
    """
    output looks like
    {
        "choices": [
                {
                "finish_reason": "stop",
                "index": 0,
                "logprobs": null,
                "text": " No.]"
                }
            ],
        "created": 1682393599,
        "id": "cmpl-793iZfhYTA2pAHA3LJSGmy6aUsQO7",
        "model": "text-davinci-003",
        "object": "text_completion",
        "usage": {
            "completion_tokens": 2,
            "prompt_tokens": 937,
            "total_tokens": 939
        }
    }
    """
    if kwargs["api_type"] == "open_ai":
        kwargs["model"] = kwargs["engine"]
        kwargs.pop("engine", None)
        kwargs.pop("api_version", None)
    # print("api_base = ", kwargs["api_base"])
    ret = openai.Completion.create(**kwargs)

    # reorder outputs to have the same order as inputs
    choices_in_correct_order = [{}] * len(ret.choices)
    for choice in ret.choices:
        choices_in_correct_order[choice.index] = choice
    ret.choices = choices_in_correct_order

    return ret


def together_completion(kwargs):
    """
    Together.ai doesn't support batch inference or async calls, so we use multithreading to simulate the behavior of batch inference.
    We use multithreading (instead of multiprocessing) because this method is I/O-bound, mostly waiting for an HTTP response to come back.
    """
    kwargs["model"] = kwargs["engine"]
    engine = kwargs.pop("engine", None)
    kwargs.pop("api_version", None)
    kwargs.pop("api_type", None)
    kwargs.pop("api_key", None)

    max_tokens = kwargs["max_tokens"]
    temperature = kwargs["temperature"]
    top_p = kwargs["top_p"]
    stop = kwargs["stop"]
    if stop is None:
        stop = []

    stop = list(stop) + ["<bot>", "<human>", "=====", "-----"]

    if (
        "presence_penalty" in kwargs
        and kwargs["presence_penalty"] is not None
        and kwargs["presence_penalty"] != 0
    ):
        logger.warning(
            "Ignoring `presence_penalty` since it is not supported by this model."
        )
    if (
        "frequency_penalty" in kwargs
        and kwargs["frequency_penalty"] is not None
        and kwargs["frequency_penalty"] != 0
    ):
        logger.warning(
            "Ignoring `frequency_penalty` since it is not supported by this model."
        )

    with ThreadPoolExecutor(
        1
    ) as executor:  # Together has an extremely low rate limit when using the free tier
        thread_outputs = [
            executor.submit(
                together.Complete.create,
                model=engine,
                prompt=p,
                temperature=temperature,
                max_tokens=max_tokens,
                top_p=top_p,
                stop=stop,
            )
            for p in kwargs["prompt"]
        ]
    thread_outputs = [o.result() for o in thread_outputs]
    thread_outputs = [
        o["output"]["choices"][0]["text"] for o in thread_outputs
    ]

    ret = {
        "choices": [
            {
                "finish_reason": "stop",
                "index": i,
                "logprobs": None,
                "text": thread_outputs[i],
            }
            for i in range(len(thread_outputs))
        ],
        "created": 0,
        "id": "",
        "model": engine,
        "object": "text_completion",
        "usage": {
            "completion_tokens": 0,
            "prompt_tokens": 0,
            "total_tokens": 0,
        },
    }

    return ret


@lru_cache_if_zero_temperature(maxsize=10000)
@retry(
    retry=retry_if_exception_type(
        (OpenAIError, together.error.RateLimitError)
    ),
    wait=wait_exponential(min=1, max=120, exp_base=2),
    stop=stop_after_attempt(7),
    before_sleep=before_sleep_log(logger, logging.ERROR),
)
def _llm_completion_with_backoff_and_cache(original_engine_name, **kwargs):
    try:
        if original_engine_name in global_variables.openai_chat_model_list:
            return asyncio.run(openai_chat_completion(kwargs))
        elif original_engine_name in global_variables.local_model_list:
            return global_variables.local_engine_clients[
                kwargs["api_base"]
            ].completion(kwargs)
        elif original_engine_name in global_variables.together_model_list:
            return together_completion(kwargs)
        elif (
            original_engine_name in global_variables.openai_nonchat_model_list
        ):
            return openai_nonchat_completion(kwargs)
        else:
            raise ValueError("This engine is not supported yet.")
    except OpenAIError as e:
        if "Azure OpenAI’s content management policy" in str(e):
            logger.error("No output due to Azure's content filtering.")
            return {
                "choices": [
                    {
                        "finish_reason": "content_filter",
                        "index": i,
                        "logprobs": None,
                        "text": None,
                    }
                    for i in range(len(kwargs["prompt"]))
                ],
                "usage": {
                    "completion_tokens": 0,
                    "prompt_tokens": 0,
                    "total_tokens": 0,
                },
            }
        else:
            raise e


def _set_llm_resource_fields(llm_resource, **kwargs):
    kwargs_copy = copy.deepcopy(kwargs)

    if "api_version" in llm_resource:
        kwargs_copy["api_version"] = llm_resource["api_version"]
    if "api_base" in llm_resource:
        kwargs_copy["api_base"] = llm_resource["api_base"]

    if "api_key" in llm_resource:
        kwargs_copy["api_key"] = llm_resource["api_key"]

    if llm_resource["api_type"] == "together":
        together.api_key = llm_resource["api_key"]

    # all providers have these fields
    kwargs_copy["api_type"] = llm_resource["api_type"]
    kwargs_copy["engine"] = llm_resource["engine_map"][kwargs["engine"]]

    # print("setting api_base to ", kwargs_copy["api_base"])

    return kwargs_copy


def _postprocess_generations(generation_output: str) -> str:
    """
    Might output an empty string if generation is not at least one full sentence
    """
    # replace all whitespaces with a single space
    generation_output = " ".join(generation_output.split())
    # print("generation_output = ", generation_output)

    original_generation_output = generation_output
    # remove extra dialog turns, if any
    turn_indicators = [
        "You:",
        "They:",
        "Context:",
        "You said:",
        "They said:",
        "Assistant:",
        "Chatbot:",
    ]
    for t in turn_indicators:
        while generation_output.find(t) > 0:
            generation_output = generation_output[: generation_output.find(t)]

    generation_output = generation_output.strip()
    # delete half sentences
    if len(generation_output) == 0:
        logger.error(
            "LLM output is empty after postprocessing. Before postprocessing it was %s",
            original_generation_output,
        )
        return generation_output

    if (
        generation_output[-1] not in {".", "!", "?"}
        and generation_output[-2:] != '."'
    ):
        # handle preiod inside closing quotation
        last_sentence_end = max(
            generation_output.rfind("."),
            generation_output.rfind("!"),
            generation_output.rfind("?"),
            generation_output.rfind('."') + 1,
        )
        if last_sentence_end > 0:
            generation_output = generation_output[: last_sentence_end + 1]
    return generation_output


def llm_generate(
    template_file: str,
    prompt_parameter_values: Union[dict, List[dict]],
    engine: str,
    max_tokens: int,
    temperature: float,
    stop_tokens,
    top_p: float = 0.9,
    frequency_penalty: float = 0,
    presence_penalty: float = 0,
    postprocess: bool = True,
    filled_prompt=None,
):
    """
    Generates continuations for one or more prompts in parallel
    Inputs:
        prompt_parameter_values: dict or list of dict. If the input is a list, the output will be a list as well
        filled_prompt: gives direct access to the underlying model, without having to load a prompt template from a .prompt file. Used for testing.
    """
    if not (
        filled_prompt is None
        and prompt_parameter_values is not None
        and template_file is not None
    ) and not (
        filled_prompt is not None
        and prompt_parameter_values is None
        and template_file is None
    ):
        raise ValueError(
            "Can only use filled_prompt if template_file and prompt_parameter_values are None"
        )

    # Decide which LLM resource to send this request to.
    # Use hash so that each time this function gets called with the same parameters after a backoff, the request gets sent to the same resource
    potential_llm_resources = [
        resource
        for resource in global_variables.all_llm_endpoints
        if engine in resource["engine_map"]
    ]
    llm_resource = potential_llm_resources[
        hash(
            str(
                (
                    template_file,
                    prompt_parameter_values,
                    engine,
                    max_tokens,
                    temperature,
                    stop_tokens,
                    top_p,
                    frequency_penalty,
                    presence_penalty,
                )
            )
        )
        % len(potential_llm_resources)
    ]
    # uniform load balancing instead of hashing
    # llm_resource = potential_llm_resources[random.randrange(len(potential_llm_resources))]

    if llm_resource["api_type"] == "local":
        prompt_format = llm_resource["prompt_format"]
    else:
        prompt_format = "none"

    # convert to a single element list so that the rest of the code only has to deal with a list
    input_was_list = True
    if filled_prompt is None:
        assert prompt_parameter_values is not None
        if not isinstance(prompt_parameter_values, list):
            input_was_list = False
            prompt_parameter_values = [prompt_parameter_values]
        filled_prompt, rendered_blocks = _fill_prompt(
            template_file, prompt_parameter_values, engine, prompt_format
        )
    else:
        if not isinstance(filled_prompt, list):
            input_was_list = False
            filled_prompt = [filled_prompt]

    assert isinstance(filled_prompt, list)
    # Call LLM to generate outputs
    generation_output = _llm_completion_with_backoff_and_cache(
        original_engine_name=engine,
        **_set_llm_resource_fields(
            llm_resource=llm_resource,
            engine=engine,
            prompt=filled_prompt,
            max_tokens=max_tokens,
            temperature=temperature,
            top_p=top_p,
            frequency_penalty=frequency_penalty,
            presence_penalty=presence_penalty,
            stop=stop_tokens,
        )
    )
    outputs = []
    for choice in generation_output["choices"]:
        if choice["text"]:
            outputs.append(choice["text"])

    logger.info(
        "LLM output: %s", json.dumps(outputs, indent=2, ensure_ascii=False)
    )

    # calculate and record the cost
    cost_prompt, cost_completion = global_variables._model_name_to_cost(engine)
    total_cost = (
        generation_output["usage"]["prompt_tokens"] * cost_prompt
        + generation_output["usage"].get("completion_tokens", 0)
        * cost_completion
    ) / 1000
    global_variables.add_to_total_cost(total_cost)

    # postprocess the generation outputs
    outputs = [o.strip() for o in outputs]
    if postprocess:
        outputs = [_postprocess_generations(o) for o in outputs]

    # add to prompt logs if needed
    if global_variables.debug_prompts:
        with global_variables.thread_lock:
            for i, o in enumerate(outputs):
                if (
                    template_file
                    in global_variables.prompts_to_skip_for_debugging
                ):
                    continue
                global_variables.prompt_logs.append(
                    {
                        "template_name": template_file,
                        "instruction": rendered_blocks[i]["short_instruction"]
                        if "short_instruction" in rendered_blocks[i]
                        else rendered_blocks[i]["instruction"],
                        "input": rendered_blocks[i]["input"],
                        "output": o,
                    }
                )

    if outputs == []:
        outputs = ""

    # convert back to a single item
    if len(outputs) == 1 and not input_was_list:
        outputs = outputs[0]
    return outputs



from collections import OrderedDict
from typing import List


class DialogueTurn:
    def __init__(
        self,
        agent_utterance: str = None,
        user_utterance: str = None,
        pipeline: str = None,
        engine: str = None,
        generate_engine: str = None,
        draft_engine: str = None,
    ):
        self.engine = engine
        self.generate_engine = generate_engine
        self.draft_engine = draft_engine
        self.pipeline = pipeline
        self.wall_time_seconds = (
            0  # how much time it took to generate this turn, in seconds
        )
        self.agent_utterance = agent_utterance
        self.user_utterance = user_utterance

        # retrieve_and_generate pipeline
        self.initial_search_query = None
        self.initial_search_query_time = None
        self.initial_search_results = []
        self.initial_search_result_titles = []
        self.initial_search_bullets = []

        # generate_and_correct pipeline
        self.llm_utterance = None
        self.claims = []
        self.verification_retrieval_results = {}
        self.verification_result = {}

        # early_combine pipeline
        self.combined_evidences = []
        self.combined_utterance = None
        self.feedback = []
        self.feedback_scores = []
        self.refined_utterance = None

    def _summarize_vc_log(self):
        verification_summary = {}
        assert len(self.verification_result) == len(
            self.verification_retrieval_results
        ), "We need to have retrieved evidence for all claims"
        for key, value in self.verification_retrieval_results.items():
            claim_idx = int(key)
            v_ret_results = []
            for v in value:
                title, paragraph, score = tuple(v)
                v_ret_results.append(
                    {
                        "title": title,
                        "paragraph": paragraph,
                        "score": round(score, 1),
                    }
                )
            verification_summary[self.claims[claim_idx][0]] = OrderedDict(
                {
                    "label": self.verification_result[claim_idx]["label"],
                    "fixed_claim": self.verification_result[claim_idx][
                        "fixed_claim"
                    ],
                    "retrieval_results": v_ret_results,
                }
            )
        return verification_summary

    def _summarize_rg_log(self):
        rg_summary = {
            "initial_search_query": self.initial_search_query,
            "initial_search_query_time": self.initial_search_query_time,
            "initial_search_bullets": self.initial_search_bullets,
            "initial_search_results": [],
        }

        for i in range(len(self.initial_search_results)):
            rg_summary["initial_search_results"].append(
                {
                    "title": self.initial_search_result_titles[i],
                    "paragraph": self.initial_search_results[i],
                    # 'bullets': self.initial_search_bullets,
                }
            )

        return rg_summary

    def log(self):
        """
        Returns a json object that contains all information inside `self`
        """
        # combine fields into a more human-readable field
        verification_summary = self._summarize_vc_log()
        rg_summary = self._summarize_rg_log()

        return OrderedDict(
            {
                # retrieve_and_generate pipeline
                "retrieve_and_generate": rg_summary,
                # generate_and_correct pipeline
                "llm_utterance": self.llm_utterance,
                "generate_and_correct": verification_summary,
                # early_combine pipeline
                "combined_evidences": self.combined_evidences,
                "combined_utterance": self.combined_utterance,
                "feedback": self.feedback,
                "feedback_scores": self.feedback_scores,
                "refined_utterance": self.refined_utterance,
                "user_utterance": self.user_utterance,
                "agent_utterance": self.agent_utterance,
                "engine": self.engine,
                "generate_engine": self.generate_engine,
                "draft_engine": self.draft_engine,
                "pipeline": self.pipeline,
                "wall_time_seconds": round(self.wall_time_seconds, 1),
            }
        )

    @staticmethod
    def utterance_list_to_dialog_history(utterance_list: List[str]):
        """
        The resulting dialog history will not have all the fields correctly initialized, since no information about e.g. search queries is available
        """
        dialog_history = []
        assert (
            len(utterance_list) % 2 == 1
        ), "The first turn is always the user, and the turn to be generated is always the agent, so the number of turns should be odd"
        for i in range(0, len(utterance_list) - 2, 2):
            dialog_history.append(
                DialogueTurn(
                    user_utterance=utterance_list[i],
                    agent_utterance=utterance_list[i + 1],
                )
            )
        user_utterance = utterance_list[-1]

        return dialog_history, user_utterance

    @staticmethod
    def dialog_history_to_utterance_list(dialog_history) -> List[str]:
        """
        Convert a list of DialogueTurns to a list of strings
        """
        utterance_list = []
        for turn in dialog_history:
            utterance_list.append(turn.user_utterance)
            utterance_list.append(turn.agent_utterance)
        return utterance_list



import logging
import re
import time
from concurrent.futures import ThreadPoolExecutor
from typing import List

import numpy as np
import requests

from agent_search import SciPhi
from agent_search.llm.llm_generate import llm_generate

from .claim_splitter import ClaimSplitter
from .dialog_turn import DialogueTurn
from .refiner import Refiner
from .utils import extract_year, is_everything_verified

logger = logging.getLogger(__name__)


class Chatbot:
    """
    A stateless chatbot. Stateless means that it does not store the history of the dialog in itself, but requires it as an input
    """

    def __init__(self, args) -> None:
        # Initialize everything, because we can change the pipeline on the fly using system_parameters
        self.claim_splitter = ClaimSplitter(args.claim_prompt_template_file)
        self.evi_num = args.evi_num
        self.colbert_endpoint = args.colbert_endpoint
        self.retrieval_num = args.retrieval_num
        self.refiner = Refiner(prompt=args.refinement_prompt, args=args)

        self.temperature = args.temperature
        self.max_tokens = args.max_tokens
        self.top_p = args.top_p
        self.presence_penalty = args.presence_penalty
        self.frequency_penalty = args.frequency_penalty
        self.skip_verification = args.skip_verification

        # default parameters, can be overridden:
        self.engine = args.engine
        self.generate_engine = args.generate_engine
        self.draft_engine = args.draft_engine
        self.do_refine = args.do_refine
        self.fuse_claim_splitting = args.fuse_claim_splitting
        self.client = SciPhi()

    def generate_next_turn(
        self,
        object_dlg_history: List[DialogueTurn],
        new_user_utterance: str,
        pipeline: str,
        system_parameters: dict = {},
    ):
        """
        Generate the next turn of the dialog
        system_parameters: can override some of the default parameters defined in __init__()
        """
        # throw error if system_parameters contains keys that are not supported
        for key in system_parameters:
            assert key in [
                "engine",
                "generate_engine",
                "draft_engine",
                "fuse_claim_splitting",
                "do_refine",
            ], f"Unsupported system_parameter key: {key}"

        engine = system_parameters.get("engine", self.engine)
        generate_engine = system_parameters.get(
            "generate_engine", self.generate_engine
        )
        if generate_engine is None:
            # this means that the default `generate_engine` was not provided via commandline, and system_parameters is not override it either.
            # So default to `engine`
            generate_engine = engine
        draft_engine = system_parameters.get("draft_engine", self.draft_engine)
        if draft_engine is None:
            draft_engine = engine
        fuse_claim_splitting = system_parameters.get(
            "fuse_claim_splitting", self.fuse_claim_splitting
        )
        engine_dict = {
            "default": engine,
            "generate": generate_engine,
            "draft": draft_engine,
        }
        do_refine = system_parameters.get("do_refine", self.do_refine)

        start_time = time.time()

        if pipeline == "generate_and_correct":
            new_dlg_turn = self.generate_and_correct_pipeline(
                object_dlg_history,
                new_user_utterance=new_user_utterance,
                engine_dict=engine_dict,
            )
        elif pipeline == "retrieve_and_generate":
            new_dlg_turn = self.retrieve_and_generate_pipeline(
                object_dlg_history,
                new_user_utterance=new_user_utterance,
                engine_dict=engine_dict,
            )
        elif pipeline == "generate":
            reply = self._generate_only(
                "baseline_chatbot.prompt",
                object_dlg_history,
                new_user_utterance=new_user_utterance,
                engine_dict=engine_dict,
            )
            new_dlg_turn = DialogueTurn(user_utterance=new_user_utterance)
            new_dlg_turn.llm_utterance = reply
            new_dlg_turn.agent_utterance = reply
        elif pipeline == "retrieve_only":
            new_dlg_turn = self.retrieve_only_pipeline(
                object_dlg_history,
                new_user_utterance=new_user_utterance,
                engine_dict=engine_dict,
            )
        elif pipeline == "early_combine":
            new_dlg_turn = self.early_combine_pipeline(
                object_dlg_history,
                new_user_utterance=new_user_utterance,
                engine_dict=engine_dict,
                fuse_claim_splitting=fuse_claim_splitting,
            )
        else:
            raise ValueError

        if do_refine == "True" or do_refine == "true" or do_refine == True:
            do_refine = True
        else:
            do_refine = False

        if do_refine:
            prerefinement_agent_utterance = new_dlg_turn.agent_utterance
            new_dlg_turn.agent_utterance = self.refiner.set_refinement_fields(
                object_dlg_history, new_dlg_turn, engine_dict=engine_dict
            )
            if new_dlg_turn.agent_utterance == prerefinement_agent_utterance:
                logger.info("Refinement did NOT change the agent utterance")

        new_dlg_turn.engine = engine
        new_dlg_turn.generate_engine = generate_engine
        new_dlg_turn.draft_engine = draft_engine
        new_dlg_turn.pipeline = pipeline

        end_time = time.time()
        new_dlg_turn.wall_time_seconds = end_time - start_time

        return new_dlg_turn

    def retrieve_only_pipeline(
        self,
        object_dlg_history: List[DialogueTurn],
        new_user_utterance: str,
        engine_dict: dict,
    ):
        new_dlg_turn = DialogueTurn(user_utterance=new_user_utterance)
        # search based on the history of the dialog so far
        search_prompt_output = llm_generate(
            template_file="query.prompt",
            prompt_parameter_values={
                "dlg": object_dlg_history,
                "new_user_utterance": new_user_utterance,
                "force_search": True,
            },
            engine=engine_dict["default"],
            max_tokens=50,
            temperature=0.0,
            top_p=0.5,
            stop_tokens=["\n"],
            postprocess=False,
        )
        search_prompt_output = (
            "Yes. " + search_prompt_output
        )  # because we are forcing a search
        self._handle_search_prompt_output(
            search_prompt_output=search_prompt_output,
            new_dlg_turn=new_dlg_turn,
            num_paragraphs=1,
            summarize_results=False,
            engine_dict=engine_dict,
        )

        paragraph = new_dlg_turn.initial_search_results[
            0
        ]  # we only retrieve one paragraph
        title = new_dlg_turn.initial_search_result_titles[0]
        new_dlg_turn.agent_utterance = (
            'I found an article titled "' + title + '": ' + paragraph
        )
        return new_dlg_turn

    def retrieve_and_generate_pipeline(
        self,
        object_dlg_history: List[DialogueTurn],
        new_user_utterance: str,
        engine_dict: dict,
    ):
        new_dlg_turn = DialogueTurn(user_utterance=new_user_utterance)
        reply = self._retrieve_and_generate(
            object_dlg_history,
            new_user_utterance,
            new_dlg_turn,
            engine_dict=engine_dict,
        )
        new_dlg_turn.agent_utterance = reply

        return new_dlg_turn

    def generate_and_correct_pipeline(
        self,
        object_dlg_history: List[DialogueTurn],
        new_user_utterance: str,
        engine_dict: str,
    ):
        """
        Verify and correct the last turn of a given dialog using retrieved evidences
        Args:
            - `object_dlg_history` (list): previous dialog turns
            - `new_user_utterance` (str): last user utterance
        Returns:
            - `corrected_reply` (str): corrected LLM response
            - `new_dialog_turn` (DialogTurn)
        """
        original_reply = self._generate_only(
            "generate.prompt",
            object_dlg_history,
            new_user_utterance,
            engine_dict=engine_dict,
        )

        new_dlg_turn = DialogueTurn(user_utterance=new_user_utterance)
        new_dlg_turn.llm_utterance = original_reply

        new_dlg_turn.agent_utterance = self._generate_and_correct_reply(
            object_dlg_history,
            new_user_utterance,
            original_reply,
            new_dlg_turn,
            engine_dict=engine_dict,
        )

        return new_dlg_turn

    def early_combine_pipeline(
        self,
        object_dlg_history: List[DialogueTurn],
        new_user_utterance: str,
        engine_dict: str,
        fuse_claim_splitting: bool,
    ):
        new_dlg_turn = DialogueTurn(user_utterance=new_user_utterance)

        # gather evidence from two routs in parallel
        with ThreadPoolExecutor(2) as executor:
            search_summary = executor.submit(
                self._search_and_summarize,
                object_dlg_history,
                new_user_utterance,
                new_dlg_turn,
                engine_dict=engine_dict,
            )
            supported_claims = executor.submit(
                self._generate_split_and_fact_check,
                object_dlg_history,
                new_user_utterance,
                new_dlg_turn,
                engine_dict=engine_dict,
                fuse_claim_splitting=fuse_claim_splitting,
            )
        search_summary = search_summary.result()
        supported_claims = supported_claims.result()

        combined_evi = search_summary + supported_claims
        # logger.info('Combined evidences: %s', new_dlg_turn.combined_evidences)
        new_dlg_turn.combined_evidences = combined_evi

        if not combined_evi:
            logger.info("Combined evidence is empty")
        #     if new_dlg_turn.initial_search_query is None:
        #         new_dlg_turn.combined_utterance = original_reply # no search needed, so return the original chitchat response
        #     else:
        #         new_dlg_turn.combined_utterance = "Sorry, I'm not sure." # will become more conversational after refinement
        # else:
        new_dlg_turn.combined_utterance = self._reply_using_combined_evidence(
            object_dlg_history,
            new_user_utterance,
            combined_evi,
            engine_dict=engine_dict,
        )
        new_dlg_turn.agent_utterance = new_dlg_turn.combined_utterance

        return new_dlg_turn

    def _handle_search_prompt_output(
        self,
        search_prompt_output: str,
        new_dlg_turn: DialogueTurn,
        num_paragraphs,
        summarize_results: bool,
        engine_dict: dict,
    ):
        """
        Updates `new_dlg_turn` with logs
        A sample output is: Yes. You Google "James E. Webb the administrator of NASA". The year of the results is "none".]
        """
        reranking_factor = 3  # we will retrieve num_paragraphs * reranking_factor paragraphs before reranking them

        search_prompt_output = search_prompt_output.strip()
        search_pattern = (
            r'Yes\. You.*"([^"]*)".* The year of the results is "([^=]*)"\.]?'
        )
        search_match = re.match(search_pattern, search_prompt_output)

        if search_prompt_output.startswith("No"):
            # sometimes LLM outputs No. with extra explanation afterwards instead of ']', or "No search needed". So this more lax condition leads to fewer Exceptions
            logger.info("No search needed.")
        elif search_match:
            search_query = search_match.group(1)
            search_query_time = search_match.group(2)
            y = extract_year(title="", passage=search_query)
            if len(y) > 0:
                logger.info("Overriding query year")
                search_query_time = y[0]
            results = self.client.search(search_query, "agent-search")
            paragraphs = [ele["text"] for ele in results]
            scores = [ele["score"] for ele in results]
            titles = [ele["title"] for ele in results]


            if summarize_results:
                bullets = []
                not_none_paragraphs = []
                not_none_titles = []
                # summarize in batches, until we reach `num_paragraphs` paragraphs that are deemed relevant
                for start_idx in range(
                    0, num_paragraphs * reranking_factor, num_paragraphs
                ):
                    b, not_none_paragraph_indices = self._summarize_results(
                        search_query,
                        paragraphs[start_idx : start_idx + num_paragraphs],
                        titles[start_idx : start_idx + num_paragraphs],
                        maximum_paragraphs_needed=num_paragraphs
                        - len(not_none_paragraphs),
                        engine_dict=engine_dict,
                    )
                    # print("not_none_paragraph_indices = ", not_none_paragraph_indices)
                    not_none_paragraphs += [
                        paragraphs[start_idx + i]
                        for i in not_none_paragraph_indices
                    ]
                    not_none_titles += [
                        titles[start_idx + i]
                        for i in not_none_paragraph_indices
                    ]
                    bullets = bullets + b
                    assert len(not_none_paragraphs) <= num_paragraphs
                    if len(not_none_paragraphs) == num_paragraphs:
                        break
                titles = not_none_titles
                paragraphs = not_none_paragraphs

            else:
                paragraphs = paragraphs[:num_paragraphs]
                titles = titles[:num_paragraphs]
                bullets = None

            # log everything
            new_dlg_turn.initial_search_query = search_query
            new_dlg_turn.initial_search_query_time = search_query_time
            new_dlg_turn.initial_search_results = paragraphs
            new_dlg_turn.initial_search_result_titles = titles
            new_dlg_turn.initial_search_bullets = bullets
        else:
            raise ValueError(
                "Search prompt's output is invalid: %s" % search_prompt_output
            )
            # logger.error('Search prompt\'s output is invalid: %s' % search_prompt_output)

    def _summarize_results(
        self,
        search_query,
        paragraphs,
        titles,
        maximum_paragraphs_needed,
        engine_dict,
    ):
        """
        Summarizes `paragraphs` and returns the indices of at most `maximum_paragraphs_needed` paragraphs that are deemed relevant to the `query`
        """
        summaries = llm_generate(
            template_file="summarize_and_filter.prompt",
            prompt_parameter_values=[
                {"title": t, "article": p, "query": search_query}
                for (t, p) in zip(titles, paragraphs)
            ],
            engine=engine_dict["default"],
            max_tokens=200,
            temperature=0.0,
            top_p=0.5,
            stop_tokens=None,
            postprocess=False,
        )
        bullets = []
        not_none_paragraph_indices = []
        for paragraph_idx, s in enumerate(summaries):
            if s.startswith("Yes. "):
                # necessary for distilled models
                s = s[5:]
            if s.startswith("None") or s == "- None" or s == "-None":
                # skip the None paragraphs
                logger.info(
                    "This retrieved paragraphs was deemed unrelated: %s",
                    paragraphs[paragraph_idx],
                )
                continue
            not_none_paragraph_indices.append(paragraph_idx)
            for b in s.split("\n-"):
                b = b.strip()
                if len(b) == 0:
                    continue
                if not b.endswith("."):
                    # most likely a partial generation that was cut off because of max_tokens
                    continue
                bullets.append(b.strip("- "))
            if len(not_none_paragraph_indices) == maximum_paragraphs_needed:
                break

        return bullets, not_none_paragraph_indices

    def _retrieve_and_generate(
        self,
        object_dlg_history: List[DialogueTurn],
        new_user_utterance: str,
        new_dlg_turn: DialogueTurn,
        engine_dict: dict,
    ) -> str:
        """
        Retrieves related documents and generates a reply base on them, given the dialog history
        Updates `new_dlg_turn` with logs
        Returns reply
        """
        self._search_and_summarize(
            object_dlg_history, new_user_utterance, new_dlg_turn, engine_dict
        )

        reply = llm_generate(
            template_file="draft.prompt",
            prompt_parameter_values={
                "dlg": object_dlg_history,
                "last_user_utterance": new_user_utterance,
                "evidences": new_dlg_turn.initial_search_bullets,
            },
            engine=engine_dict["default"],
            max_tokens=self.max_tokens,
            temperature=self.temperature,
            top_p=self.top_p,
            presence_penalty=self.presence_penalty,
            stop_tokens=["\n"],
            postprocess=True,
        )
        return reply

    def _search_and_summarize(
        self,
        object_dlg_history: List[DialogueTurn],
        new_user_utterance: str,
        new_dlg_turn: DialogueTurn,
        engine_dict: dict,
    ):
        search_prompt_output = llm_generate(
            template_file="query.prompt",
            prompt_parameter_values={
                "dlg": object_dlg_history,
                "new_user_utterance": new_user_utterance,
                "force_search": False,
            },
            engine=engine_dict["default"],
            max_tokens=50,
            temperature=0.0,
            top_p=0.5,
            stop_tokens=["\n"],
            postprocess=False,
        )
        self._handle_search_prompt_output(
            search_prompt_output=search_prompt_output,
            new_dlg_turn=new_dlg_turn,
            num_paragraphs=self.retrieval_num,
            summarize_results=True,
            engine_dict=engine_dict,
        )
        return new_dlg_turn.initial_search_bullets

    def _generate_split_and_fact_check(
        self,
        object_dlg_history: List[DialogueTurn],
        new_user_utterance: str,
        new_dlg_turn: DialogueTurn,
        engine_dict: dict,
        fuse_claim_splitting: bool,
    ):
        original_reply = self._generate_only(
            "generate.prompt",
            object_dlg_history,
            new_user_utterance,
            engine_dict=engine_dict,
        )
        if not fuse_claim_splitting:
            new_dlg_turn.llm_utterance = original_reply
            claims_output = None
        else:
            new_dlg_turn.llm_utterance = None
            claims_output = original_reply

        claims = self.claim_splitter.split_claim(
            dialog_history=object_dlg_history,
            new_user_utterance=new_user_utterance,
            current_agent_utterance=original_reply,
            engine_dict=engine_dict,
            claims_output=claims_output,
        )

        new_dlg_turn.claims = claims
        if not claims:
            logger.info("No claims to check")
            return []

        # retrieve evidence
        ret_output = self._retrieve_evidences(claims)

        # verify claims
        ver_output = self._verify_claims(
            claims,
            ret_output,
            object_dlg_history,
            new_user_utterance,
            original_reply,
            do_correct=False,
            engine_dict=engine_dict,
        )

        new_dlg_turn.verification_retrieval_results = ret_output
        new_dlg_turn.verification_result = ver_output

        # only keep supported claim
        supported_claims = []
        for label_fix in ver_output:
            verification_label, fixed_claim = (
                label_fix["label"],
                label_fix["fixed_claim"],
            )
            if verification_label == "SUPPORTS":
                supported_claims.append(fixed_claim)
        return supported_claims

    def _generate_and_correct_reply(
        self,
        object_dlg_history: List[DialogueTurn],
        new_user_utterance: str,
        original_reply: str,
        new_dlg_turn: DialogueTurn,
        engine_dict: dict,
    ) -> str:
        """
        Verifies and corrects `original_reply` given the dialog history
        Updates `new_dlg_turn` with logs
        Returns corrected reply
        """
        # split claims
        # the returned "claims" is a list of tuples (claim, year)
        claims = self.claim_splitter.split_claim(
            dialog_history=object_dlg_history,
            new_user_utterance=new_user_utterance,
            current_agent_utterance=original_reply,
            engine_dict=engine_dict,
        )
        claims = ClaimSplitter.remove_claims_from_previous_turns(
            claims, object_dlg_history
        )
        if not claims:
            logger.info("No claims to check")
            return original_reply
        new_dlg_turn.claims = claims

        # retrieve evidence
        ret_output = self._retrieve_evidences(claims)

        # TODO: use the ret_output together with initial search outputs for verification
        # verify claims
        ver_output = self._verify_claims(
            claims,
            ret_output,
            object_dlg_history,
            new_user_utterance,
            original_reply,
            do_correct=True,
            engine_dict=engine_dict,
        )

        # update dialog turn
        new_dlg_turn.verification_retrieval_results = ret_output
        new_dlg_turn.verification_result = ver_output
        if is_everything_verified(ver_output):
            logger.info("All claims passed verification, nothing to correct")
            return original_reply

        # correction
        corrected_reply = original_reply
        fixed_claims = []
        for label_fix in ver_output:
            verification_label, fixed_claim = (
                label_fix["label"],
                label_fix["fixed_claim"],
            )
            if (
                verification_label == "SUPPORTS"
            ):  # if the claim is already correct, no need to fix
                continue
            fixed_claims.append(fixed_claim)
        assert len(fixed_claims) > 0
        corrected_reply = self._correct(
            original_reply,
            object_dlg_history,
            new_user_utterance,
            fixed_claims,  # corrected claim for REFUTE and "I'm not sure" for NOT ENOUGH INFO claims.
            engine_dict=engine_dict,
        )

        return corrected_reply

    def _generate_only(
        self,
        generation_prompt: str,
        dialog_history: List[DialogueTurn],
        new_user_utterance: str,
        engine_dict: dict,
    ) -> str:
        """
        Generate baseline LLM response
        Args:
            - `generation_prompt` (str): the .prompt file to use for this stage
            - `dialog_history` (list): previous turns
        Returns:
            - `reply`(str): original LLM response
        """
        reply = llm_generate(
            template_file=generation_prompt,
            prompt_parameter_values={
                "dlg": dialog_history,
                "new_user_utterance": new_user_utterance,
                "engine_name": engine_dict[
                    "generate"
                ],  # used to enforce model knowledge cut-off date for models other than GPT-4
            },
            engine=engine_dict["generate"],
            max_tokens=self.max_tokens,
            temperature=self.temperature,
            stop_tokens=["\n"],
            top_p=self.top_p,
            frequency_penalty=self.frequency_penalty,
            presence_penalty=self.presence_penalty,
            postprocess=True,
        )

        return reply

    def _correct(
        self,
        original_reply,
        object_dlg_history,
        last_user_utterance,
        fixed_claims,
        engine_dict: dict,
    ):
        """
        Given context + original response + evidence for a claim, fix the original response

        Args:
            - `original_reply`(str): LLM's original response
            - `object_dlg_history`(list): list of previous DialogueTurns
            - `last_user_utterance` (str): last user utterance
            - `fixed_claims` (list): list of fixed claims
        Returns:
            - `corrected_reply`(str): corrected LLM response
        """
        # correction prompt's context should be in one line
        correction_reply = llm_generate(
            template_file="correction_combiner.prompt",
            prompt_parameter_values={
                "dlg": object_dlg_history,
                "last_user_utterance": last_user_utterance,
                "original_reply": original_reply,
                "fixed_claims": fixed_claims,
            },
            engine=engine_dict["default"],
            max_tokens=self.max_tokens,
            temperature=0,
            stop_tokens=["\n"],
            top_p=self.top_p,
            frequency_penalty=self.frequency_penalty,
            presence_penalty=self.presence_penalty,
            postprocess=True,
        )

        return correction_reply

    def _reply_using_combined_evidence(
        self,
        object_dlg_history,
        last_user_utterance,
        evidences,
        engine_dict: dict,
    ):
        combined_reply = llm_generate(
            template_file="draft.prompt",
            prompt_parameter_values={
                "dlg": object_dlg_history,
                "last_user_utterance": last_user_utterance,
                "evidences": evidences,
            },
            engine=engine_dict["draft"],
            max_tokens=self.max_tokens,
            temperature=0,
            stop_tokens=None,
            top_p=self.top_p,
            frequency_penalty=self.frequency_penalty,
            presence_penalty=self.presence_penalty,
            postprocess=True,
        )

        return combined_reply

    def _colbert_retrieve(
        self,
        query: str,
        num_paragraphs: int,
        rerank="none",
        top_p=1,
    ):
        """
        Args:
            `num_paragraphs`: number of paragraphs that will be output
            `rerank` (str): one of 'none', 'recent' or a year like '2005'. 'none' disables reranking. 'recent' retrieves more and returns the most recent ones.
                            '2005' boosts the ranking of results that match 2005. The date of a result is determined by the year numbers it contains.
            `top_p` (float): chooses from the smallest possible set of results whose cumulative probability exceeds top_p
        Returns:
            `passages` (list): a list of passage texts (excluding the title) with the highest similarities to the `query`
            `passage_scores` (list): a list of similarity scores of each passage in `passsages` with `query`
            `passage_titles` (list): a list of passage titles
        """

        # print(self.colbert_endpoint, {'query': query, 'evi_num': num_paragraphs})
        response = requests.get(
            self.colbert_endpoint,
            json={"query": query, "evi_num": num_paragraphs},
        )
        if response.status_code != 200:
            raise Exception("ColBERT Search API Error: %s" % str(response))
        results = response.json()
        passages = []
        passage_titles = []
        for r in results["passages"]:
            r = r.split("|", maxsplit=1)
            passage_titles.append(r[0].strip())
            passages.append(r[1].strip())
        scores = results["passage_scores"]
        probs = results["passage_probs"]
        # print("probs = ", probs)
        top_p_cut_off = np.cumsum(probs) > top_p
        if not np.any(top_p_cut_off):
            # even if we include everything, we don't get to top_p
            top_p_cut_off = len(scores)
        else:
            top_p_cut_off = np.argmax(top_p_cut_off) + 1
        # print("top_p_cut_off = ", top_p_cut_off)
        passages, scores, passage_titles = (
            passages[:top_p_cut_off],
            scores[:top_p_cut_off],
            passage_titles[:top_p_cut_off],
        )

        if rerank == "none":
            pass
        else:
            all_passage_dates = []
            for t, p in zip(passage_titles, passages):
                passage_years = extract_year(title=t, passage=p)
                all_passage_dates.append(passage_years)
            if rerank == "recent":
                sort_fn = lambda x: max(
                    x[3] if len(x[3]) > 0 else [0]
                )  # sort based on the latest year mentioned in the paragraph, demoting paragraphs that don't mention a year
            else:
                # rerank is a year
                try:
                    query_year = int(rerank)
                except ValueError as e:
                    # raise ValueError('rerank should be none, recent or an integer.')
                    logger.error(e)
                    return (
                        passages[:num_paragraphs],
                        scores[:num_paragraphs],
                        passage_titles[:num_paragraphs],
                    )
                sort_fn = lambda x: x[3].count(
                    query_year
                )  # boost the passages that have a matching year with the query, the more they mention the date the more we boost

            # logger.info('Search result dates before date-based reranking: %s', str(all_passage_dates))
            passages, scores, passage_titles, all_passage_dates = list(
                zip(
                    *sorted(
                        zip(
                            passages, scores, passage_titles, all_passage_dates
                        ),
                        reverse=True,
                        key=sort_fn,
                    )
                )
            )
            # logger.info('Search result dates after date-based reranking: %s', str(all_passage_dates))

        # choose top num_paragraphs paragraphs
        passages, scores, passage_titles = (
            passages[:num_paragraphs],
            scores[:num_paragraphs],
            passage_titles[:num_paragraphs],
        )

        return passages, scores, passage_titles

    def _retrieve_evidences(self, claims, top_p: float = 1):
        """
        Retrieve evidences
        Args:
            - `claims` (list): list of (claim, year)
            - `top_p` (float): chooses from the smallest possible set of results whose cumulative probability exceeds top_p
        Returns:
            - `ret_output` (dict): a dict from claim_id to a list of `evidence`
                - each `evidence` is a list of length 5: [`title of wikipedia page`, `wikipedia text`, `similarity_score`]
        """
        ret_output = dict()
        for id, (cl, year) in enumerate(claims):
            results = self.client.search(cl, "agent-search")
            evidences = []
            for result in results:
                evidences.append(
                    [result["title"], result["text"], result["score"]]
                )

            ret_output[id] = evidences

        return ret_output

    def _verify_claims(
        self,
        claims,
        ret_output,
        object_dlg_history,
        new_user_utterance,
        original_reply,
        do_correct: bool,
        engine_dict: dict,
    ):
        """
        Verify claims using retrieval output
        Args:
            - `claims` (list): list of (claim, year) pairs splitted
            - `ret_output` (dict): a dict from claim_id to a list of `evidence`
                - each `evidence` is a list of length 5: [`title of wikipedia page`, `wikipedia text`, `similarity_score`]
            - `object_dlg_history`(str): list of previous DialogueTurns
            - `last_user_utterance` (str): last user utterance
            - `original_reply`(str): original LLM response
        Returns:
            - `ver_output` (list): a list of verification label ("SUPPORTS", "REFUTES", "NOT ENOUGH INFO") and the fixed claims
        """
        ver_output = []
        parameter_values_list = []

        for claim_id, (cl, year) in enumerate(claims):
            evidences = ret_output[claim_id][: self.evi_num]
            parameter_values_list.append(
                {
                    "dlg": object_dlg_history,
                    "last_user_utterance": new_user_utterance,
                    "original_reply": original_reply,
                    "claim": cl,
                    "evidence_titles": [e[0] for e in evidences],
                    "evidence_texts": [e[1] for e in evidences],
                    "do_correct": do_correct,
                }
            )

        # when using gold evidence, we do not split claim so claim is the same with original reply
        if self.skip_verification:
            all_verification_responses = ['is "SUPPORTS"'] * len(claims)
        else:
            all_verification_responses = llm_generate(
                template_file="verify.prompt",
                prompt_parameter_values=parameter_values_list,
                engine=engine_dict["default"],
                max_tokens=200,
                temperature=0,
                stop_tokens=None,
                postprocess=False,
            )

        for (cl, year), verification_response in zip(
            claims, all_verification_responses
        ):
            # logger.info("claim: %s ; verification_response: %s", cl, verification_response)
            # the following handles cases where smaller models like gpt-35-turbo do not follow the few-shot examples' format
            if (
                'is "supports"' in verification_response.lower()
                or "no fact-checking is needed for this claim"
                in verification_response.lower()
                or "the fact-checking result is not applicable to this response"
                in verification_response.lower()
            ):
                verification_label = "SUPPORTS"
                fixed_claim = cl
            elif (
                'the fact-checking result is "not enough info"'
                in verification_response.lower()
            ):
                verification_label = "NOT ENOUGH INFO"
                fixed_claim = ""
            else:
                verification_label = "REFUTES"  # default set to be "REFUTES"
                fixed_claim = ""

            if do_correct and verification_label != "SUPPORTS":
                if "You rewrite your claim:" in verification_response:
                    fixed_claim = verification_response.split(
                        "You rewrite your claim:"
                    )[-1].strip()
                else:
                    logger.error(
                        "verification prompt did not fix a %s. Output: %s"
                        % (verification_label, verification_response)
                    )

            ver_output.append(
                {"label": verification_label, "fixed_claim": fixed_claim}
            )

        return ver_output
